#include <QDebug>
#include <math.h>
#include <complex>
#include "dsp.h"
#include "lib.h"

using namespace std;
typedef complex<double> dcomplex;

#define MAX_AUDIO 	0x7fff
#ifdef WIN32
#define M_PI	3.141592653
#endif


Dsp::Dsp(int rate) {
	settings =  new QSettings("sdr","rasdr");
	sampleRate = rate;
	bufPos = 0;
	mode = MODE_LSB;
	manualGain = 1;
	agcDec = 1/sampleRate/10; 
	agcMax = 0;
	mute = true;
	phase = 0;
	sensitivity = 0.2;
	avSqrt = 0;
	volume = 1;
	micGain = 1;
	signCnt = 0;
	maxIn = 0;
	mgain = 1;
	notch = 0;
	squelch = 0; 

	in  = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * NOTCH_FFT_SIZE);
	out = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * NOTCH_FFT_SIZE);
	plan = fftwf_plan_dft_1d(NOTCH_FFT_SIZE, in, out, FFTW_FORWARD, FFTW_ESTIMATE);

	for (int i=0;i<NOTCH_FFT_SIZE;i++)
		window[i]=0.5  - 0.5  * cos (2 * M_PI / (NOTCH_FFT_SIZE) * i);  // Hanning window

	tone1Phase = 0;
	tone2Phase = 0;
	cwTonePhase = 0;
	cwTone = getSettings(settings,"Common/cwTone",600);
	cwTone = cwTone*2*M_PI/sampleRate;
	smtrAverage = getSettings(settings,"Common/smtrAverage",1000);
	txGainAttRate = 1.0+(double)getSettings(settings,"Common/txGainAttRate",100)/1e6;
	delete settings;
}

Dsp::~Dsp() {
	fftwf_destroy_plan(plan);
	fftwf_free(in); 
	fftwf_free(out);
}

void Dsp::rxAdd(double re, double im) {
	averageSqrt(re,im);
	if (mute) {
		re = 0; 
		im = 0;
	}
	double agcGain = agc(re,im);
	if (agcGain < 1.0/(1<<31))
		agcGain = 1.0/(1<<31);
	outBuf[bufPos++] = demod(re,im, manualGain>0 ? manualGain : 1/agcGain);
	if (bufPos == arraysize(outBuf))
		bufPos = 0;
}


void Dsp::setNotchFilter(double fn) {
	// Mitra, S.K., "Design of Digital Notch Filters," Communications, IEEE Transactions on , vol.22, no.7, pp.964,970,
	// Jul 1974)
	double fs = sampleRate;
	double fb = 200;
	double omega0T = fn/(fs/2)*M_PI;
	double deltaT  = fb/(fs/2)*M_PI;
	notch_a2 = (1-tan(deltaT/2))/(1+tan(deltaT/2));
	notch_a1 = (1+notch_a2)*cos(omega0T);
}

void Dsp::setNotch(int v) {
	notch = v;
}

double Dsp::processNotch1(double in) {
	static double xst1 = 0;
	static double xst2 = 0;
	static double temp1 = 0;
	static double temp2 = 0;
	static double theta = 0; // initial value for notch frequency parameter.
	static double sth = sin(theta);
	static double cth = cos(theta);
	static double bw = 0.20*M_PI; // bandwidth parameter for notch filter.
	static double sth2 = sin(bw);
	static double cth2 = cos(bw);
	static double mu = 0.043; // adaptive filter step size.

	//
	// adaptive lattice notch filter:
	//
	temp1 = cth2*in-sth2*xst2;
	temp2 = sth2*in+cth2*xst2;
	double error = mu*(in + temp2); // notch filter output times step size.
	theta = theta - error*xst1; // coefficient update.

#if 0
	double freqnew;
	if (theta>M_PI/2)
		freqnew = 0.5*(2*M_PI-(theta+M_PI/2))/M_PI; // instantaneous freq. estimate.
	else if (theta>-M_PI/2)
		freqnew = 0.5*(theta+M_PI/2)/M_PI;
	else
		freqnew = -0.5*(theta+M_PI/2)/M_PI;
	qDebug()  << freqnew*sampleRate;
#endif

	sth = sin(theta);
	cth = cos(theta);
	xst2 = sth*temp1+cth*xst1;
	xst1 = cth*temp1-sth*xst1;

	return in+temp2;
}


double Dsp::processNotch(double in) {
	static double z0=0,z1=0,z2=0;

	double b = z0 + z2;
	double a = notch_a1*z1 - notch_a2*b;

	z2 = z1;
	z1 = z0 + a;
	z0 = in;

	return b - a; 
}

void Dsp::averageSqrt(double re, double im) {
	avSqrt = (avSqrt * (smtrAverage-1) + re*re+im*im)/smtrAverage;
}

short Dsp::demod(double re, double im, double gain) {
	double v,d_re,d_im;
	static int notchFFTSample=0;
	static double lastNotchFreq;

	if (std::isnan(gain) || std::isinf(gain))
		return 0;

	if (notch==1) {
		if (notchFFTSample<NOTCH_FFT_SIZE) {
			in[notchFFTSample][0] = re*window[notchFFTSample]; 
			in[notchFFTSample][1] = im*window[notchFFTSample]; 
		}

		if (notchFFTSample == NOTCH_FFT_SIZE) {
			
			fftwf_execute(plan);

			double notchFreq = 0;
			double maxLevel = 0;
			for (int i=0;i<NOTCH_FFT_SIZE;i++) {
				double level = out[i][0]*out[i][0] + out[i][1]*out[i][1]; 
				if (maxLevel < level) {
					maxLevel = level;
					notchFreq = i;
				}
			}
			if (notchFreq > NOTCH_FFT_SIZE/2)
				notchFreq = NOTCH_FFT_SIZE-notchFreq;
			//qDebug() << notchFreq*sampleRate/NOTCH_FFT_SIZE << maxLevel;
			if (lastNotchFreq == notchFreq) {
				setNotchFilter(notchFreq*sampleRate/NOTCH_FFT_SIZE);
			}
			lastNotchFreq = notchFreq;
		}
		if (notchFFTSample++ > NOTCH_FFT_SIZE) 
		//if (notchFFTSample++ > sampleRate) 
			notchFFTSample=0;
	}

	if (avSqrt<squelch) {
		re=0;
		im=0;
	}

	switch (mode) {
		case MODE_CW:
			{
				cwTonePhase += cwTone; 
				while (cwTonePhase > M_PI)
					cwTonePhase -= (2*M_PI);
				double re1 = sin(cwTonePhase);
				double im1 = cos(cwTonePhase);
				double v = (re*re1-im*im1)+(re*im1+im*re1);
#if 0
				// frequenzmessung, cwTone ist auf 0 zu setzen
				if (v>0 && old_v<0) {
					qDebug() << "FreqDiff:" << sampleRate/signCnt << "Hz";
					signCnt=0;
				}
				old_v = v;
				signCnt++;
#endif
				return double2short(v*gain*volume);
			}
		case MODE_LSB:
		case MODE_USB:
			if (notch==1)
				return double2short(processNotch((re+im)*gain)*volume);
			if (notch==2)
				return double2short(processNotch1((re+im)*gain)*volume);
			return double2short((re+im)*gain*volume);
		case MODE_AM:
			return double2short(sqrt(re*re + im*im)*gain*volume);
		case MODE_FM:
			v = sqrt(re*re+im*im);
			re = re / v;
			im = im / v;
			d_re = re - re_l;
			d_im = im - im_l;
			re_l = re;
			im_l = im;
			return double2short(asin(d_im*re-d_re*im)*volume);
	}
	return 0;
}

int Dsp::mod(short in, unsigned char *out) {
	short re=0,im=0;
	int in1; 
	static int last_in;

	if (micGain > 0) 
		in1 = micGain*in;
	else {
		in1 = mgain*in;
		if (abs(in1)>32000) {
			mgain = mgain * 32000.0/abs(in1);
			in1 = mgain*in;
		} else {
			mgain *= txGainAttRate;
		}
		// qDebug() << mgain << in1;
	}

	if (maxIn>0)
		maxIn = maxIn - 1;
	if (maxIn < abs(in1)) 
		maxIn = abs(in1);


	switch (mode) {
		case MODE_LSB:
		case MODE_USB:
			re = checkRange(in1, -MAX_AUDIO, MAX_AUDIO);
			break;
		case MODE_AM:
			re = checkRange(in1/2 + MAX_AUDIO/4, -MAX_AUDIO, MAX_AUDIO);
			break;
		case MODE_FM:
			phase += sensitivity * short2double(in1-last_in); 
			last_in = in1;
			while (phase > M_PI)
				phase -= (2*M_PI);
			while (phase < -M_PI)
				phase += (2*M_PI);
			re = double2short(sin(phase)/2);  
			im = double2short(cos(phase)/2);  
			//qDebug() << phase << sin(phase) << cos(phase)  << re << im;
			break;
	}
	*(out+0) = re & 0xff;;
	*(out+1) = re >> 8;
	*(out+2) = im & 0xff;
	*(out+3) = im >> 8;

	return maxIn;
}

short Dsp::tone() {
	tone1Phase += tone1Freq; 
	tone2Phase += tone2Freq; 
	while (tone1Phase > M_PI)
		tone1Phase -= (2*M_PI);
	while (tone2Phase > M_PI)
		tone2Phase -= (2*M_PI);
	return double2short(sin(tone1Phase)*tone1Level+sin(tone2Phase)*tone2Level);
}

void Dsp::setTestTone(double f1, double level1, double f2, double level2) {
	tone1Freq = f1/sampleRate*2*M_PI;
	tone2Freq = f2/sampleRate*2*M_PI;
	tone1Level = level1;
	tone2Level = level2;
}

short Dsp::double2short(double v) {
	int r = v*MAX_AUDIO;
	return checkRange(r, -MAX_AUDIO, MAX_AUDIO);
}

double Dsp::short2double(short v) {
	double r = v;
	return r/MAX_AUDIO;
}

double Dsp::agc(double re, double im) {
	double v = fabs(re+im);
	agcMax *= (1-agcDec);
	if (v>agcMax) 
		agcMax = v;
	return agcMax;
}

void Dsp::setGain(double v) {
	manualGain = v;
}

void Dsp::setAGCDec(double v) {
	agcDec = v;
}

void Dsp::muteSignal(bool v) {
	mute = v;
}

void Dsp::setSampleRate(int v) {
	sampleRate = v;
}

void Dsp::setMode(int v) {
	mode = v;
}

void Dsp::bufReset() {
	bufPos = 0;
}

double Dsp::getSmtr() {
	return avSqrt;
}

double Dsp::getAgc() {
	return agcMax;
}

void Dsp::setVolume(double v) {
	volume = v;
}

void Dsp::setMicGain(double v) {
	micGain = v;
}

void Dsp::setSquelch(int level) {
	squelch = exp((double)level/5)/exp(256/5);
}
