#include <QByteArray>
#include <QFile>
#include "assert.h"
#include <unistd.h>
#include <stdint.h>
#include "config.h"
#include "hiqsdr.h"
#include "lib.h"

#define NBITS				32

#define COEF_LEN	5
#define COEF_CNT	4

#define FILTER_TIMEOUT 	100
#define FFT_TIMEOUT 	200

#undef CREATE_MIF

Hiqsdr::Hiqsdr (QString ip, int rate) {

	sampleRate = rate;
	settings =  new QSettings("sdr","rasdr");
	active = false;
	audioOff = true;
	hiqsdrIP = ip;
	rawAddr = 0;
	rawSrc = RAW_SRC_FULL;
	txPos = 0;
	rawRead = true;
	udpBuf[txPos++]=0;
	udpBuf[txPos++]=0;
	//hiqsdrClock = getSettings(settings,"Common/hiqsdrClock",122879580); // my clock
	hiqsdrClock = getSettings(settings,"Common/hiqsdrClock",RX_CLOCK);
	smtrOffset = getSettings(settings,"Common/smtrOffset",0);
	delete settings;

	dsp = new Dsp(sampleRate);
	dsp->setNotchFilter(1000);

	ctlSocket = new QUdpSocket(this);
	ctlSocket->bind(RASDR_CTL_PORT);
	connect(ctlSocket, SIGNAL(readyRead()), this, SLOT(readCtlHiqsdr()),Qt::QueuedConnection);

	rawHiqSocket = new QUdpSocket(this);
	rawHiqSocket->bind(RASDR_RAW_PORT);
	connect(rawHiqSocket, SIGNAL(readyRead()), this, SLOT(readRaw()));

	rxSocket = new QUdpSocket(this);
	rxSocket->bind(RASDR_RX_PORT);
	connect(rxSocket, SIGNAL(readyRead()), this, SLOT(readHiqsdr()));
	
	txSocket = new QUdpSocket(this);
	txSocket->bind(RASDR_TX_PORT);

	agcTimer = new QTimer(this);
	connect(agcTimer, SIGNAL(timeout()), this, SLOT(agcTimeout()));
	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(timeout()));
	muteTimer = new QTimer(this);
	connect(muteTimer, SIGNAL(timeout()), this, SLOT(muteTimeout()));
	filterTimer = new QTimer(this);
	connect(filterTimer, SIGNAL(timeout()), this, SLOT(filterTimeout()));
	fftTimer = new QTimer(this);
	connect(fftTimer, SIGNAL(timeout()), this, SLOT(fftTimeout()));

	rxFreq = 7100000;
	txFreq = rxFreq;
	ant = 0;
	presel = 0;
	att = 0;
	txLevel = 30;
	txRate = sampleRate;
	ptt = false;
	fdx = false;
	toneTest = true ;
	txBand = getBand(txFreq);
	rxBand = getBand(rxFreq);
	txDel = 0;
	lastTX = 0;
	preamp = 0;
	
	fftTime = 0;
	fftSize = FFT_SIZE;
	byteMode = 3;
	cwMode = false;
	firmVersion = 1; 
	memset(ctlBuf, 0, sizeof(ctlBuf));
	strncpy((char*)ctlBuf, "St", 2);
	ctlBuf[13]=firmVersion;

	for (int i=0;i<MAX_ENCODERS;i++) {
		lastStep[i] = 0;
		initStep[i] = true;
	}
	setNBLevel(255);
	setPresel(presel);
	setRXFreq(rxFreq);
	setTXFreq(txFreq);
	setSampleRate(sampleRate);
	setAnt(ant);
	setAttenuator(att);
	setTXLevel(txLevel);
	setTXRate(txRate);
	setPtt(ptt);
	agcTimer->start(200);
	dsp->setTestTone(1111,0.5,1711,0.5);
	it = coef.begin();
	setFilterReady = true;
    hiqRun = true;
}


Hiqsdr::~Hiqsdr() {
	stopRX();
	delete dsp;
	delete timer;
	delete agcTimer;
	delete muteTimer;
	delete filterTimer;
	delete txSocket;
	delete rxSocket;
	delete rawHiqSocket;
	delete ctlSocket;
}


void Hiqsdr::startRX() {
	if (!active)
		return;
	datagram.clear();
	datagram.append(QString("rr"));
	rxSocket->writeDatagram(datagram, QHostAddress(hiqsdrIP), HIQSDR_RX_PORT);
	ctlSocket->writeDatagram("IO\x7f\x00\x00", 5, QHostAddress(hiqsdrIP), HIQSDR_CTL_PORT);
}

void Hiqsdr::stopRX() {
	if (!active)
		return;
	datagram.clear();
	datagram.append(QString("ss"));
	rxSocket->writeDatagram(datagram, QHostAddress(hiqsdrIP), HIQSDR_RX_PORT);
}

void Hiqsdr::hiqSend() {
	if (!active)
		return;

	if (ctlBuf[13]>0)
		ctlSocket->writeDatagram(QByteArray((const char*)ctlBuf, 24), QHostAddress(hiqsdrIP), HIQSDR_CTL_PORT);
	else
		ctlSocket->writeDatagram(QByteArray((const char*)ctlBuf, 14), QHostAddress(hiqsdrIP), HIQSDR_CTL_PORT);
}

void Hiqsdr::setRXFreq(int f) {
	rxFreq = f;
	rxBand = getBand(f);

	f = f % hiqsdrClock;
	if (f > hiqsdrClock/2)
		f = hiqsdrClock - f;

	unsigned int ph = ((int)((double)(f) / hiqsdrClock*0x100000000LL+0.5)) & 0xffffffff;

	ctlBuf[2] = (ph >> 0 ) & 0xff; 
	ctlBuf[3] = (ph >> 8 ) & 0xff; 
	ctlBuf[4] = (ph >> 16) & 0xff; 
	ctlBuf[5] = (ph >> 24) & 0xff; 

	setPresel(rxBand);
}

void Hiqsdr::setTXFreq(int f) {
	txFreq = f;
	txBand = getBand(f);

	unsigned int ph = ((int)((double)(f) / hiqsdrClock*0x100000000LL+0.5)) & 0xffffffff;

	ctlBuf[6] = (ph >> 0 ) & 0xff; 
	ctlBuf[7] = (ph >> 8 ) & 0xff; 
	ctlBuf[8] = (ph >> 16) & 0xff; 
	ctlBuf[9] = (ph >> 24) & 0xff; 

	setPresel(txBand);
	setTXLevel(txLevel);
}

void Hiqsdr::setSampleRate(int rate) {
	sampleRate = rate;
	byteMode = 3;
	switch (rate) {
		case 8000:
		case 9600:
		case 12000:
		case 16000:
		case 19200:
		case 24000:
		case 38400:
			ctlBuf[12] = (2<<6) | ((RX_CLOCK / 320 / rate) - 1);
			hiqSend();
			break;
		case 48000:
		case 60000:
		case 96000:
		case 120000:
		case 192000:
		case 240000:
		case 320000:
		case 384000:
		case 480000:
		case 640000:
		case 960000:
			ctlBuf[12] = (RX_CLOCK / 64 / rate) - 1;
			hiqSend();
			break;
		case 1280000:
		case 1536000:
		case 1920000:
			byteMode = 2;
			ctlBuf[12] = (1<<6) | ((RX_CLOCK / 16 / rate) - 1);
			hiqSend();
			break;
		case 2560000:
		case 3840000:
			byteMode = 1;
			ctlBuf[12] = (3<<6) | ((RX_CLOCK / 16 / rate) - 1);
			hiqSend();
			break;
		default:
			;
	}
	dsp->setSampleRate(rate);
}

void Hiqsdr::setFdx(bool on) {
	fdx = on;
}

void Hiqsdr::setPtt(bool on) {
	ptt = on;
	ctlBuf[11] &= 0xf0;
	ctlBuf[11] |= (cwMode ? TX_CW: TX_SSB); 
	if (on) 
		ctlBuf[11] |=  TX_PTT; 
	if (firmVersion>0) 
		ctlBuf[11] |= USE_EXT_IO; 
	hiqSend();

	if (!fdx) {
		if (ptt) {
			muteTimer->stop();
			dsp->muteSignal(true);
		} else
			muteTimer->start(300);
	}
}

void Hiqsdr::setCWMode(bool cw) {
	ctlBuf[11] &= 0xf0;
	cwMode = cw;
	ctlBuf[11] |= (cwMode ? TX_CW: TX_SSB); 
	ctlBuf[20] = (cwMode ? txDel: 0); 
	if (firmVersion>0) 
		ctlBuf[11] |= USE_EXT_IO; 
	hiqSend();
}


void Hiqsdr::setTXRate(int rate) {
	ctlBuf[11] &= 0x0f;
	switch (rate) {
		case 192000:
			ctlBuf[11] |= 0x10;
			break;
		case 480000:
			ctlBuf[11] |= 0x20;
			break;
		case 8000:
			ctlBuf[11] |= 0x30;
			break;
		default:
			;
	}
	hiqSend();
}

void Hiqsdr::setTXLevel(int l) {
	txLevel = l;
	ctlBuf[10] = (txLevel*bandTxLevel[txBand])>>8;	
	hiqSend();
}

void Hiqsdr::setAttenuator(int n) {
	ctlBuf[15]=0;
	preamp = 0;
	if (n<0) {
		preamp = 1;
		setRXFreq(rxFreq);
		return;
	}
	if (n>=20) {
		n-=20;
		ctlBuf[15] |= 0x10;
	}
	if (n>=10) {
		n-=10;
		ctlBuf[15] |= 0x08;
	}
	if (n>=8) {
		n-=8;
		ctlBuf[15] |= 0x04;
	}
	if (n>=4) {
		n-=4;
		ctlBuf[15] |= 0x02;
	}
	if (n>=2) {
		n-=2;
		ctlBuf[15] |= 0x01;
	}
	setRXFreq(rxFreq);
}

void Hiqsdr::setPresel(int n) {

	ctlBuf[14] = n | (preamp<<4); 
	hiqSend();
}

void Hiqsdr::setAnt(bool on) {
	if (on)
		ctlBuf[16] |= HIQ_ANT; 
	else
		ctlBuf[16] &= ~HIQ_ANT; 
	hiqSend();
}

void Hiqsdr::setTxDelay(int txDelay) {
	txDel = txDelay;
	if (cwMode)
		ctlBuf[20] = (txDel <<4) & 0xff; 
	hiqSend();
}

void Hiqsdr::setNBLevel(int level) {
	nbLevel = level;
	level = exp((double)level/21.0);
	if (level>0xffff)
		level=0xffff;
	//qDebug() << "level" << level;
	ctlBuf[22] = level & 0xff;
	ctlBuf[23] = level >> 8;
	hiqSend();
}

void Hiqsdr::setFilter(std::vector<gr_complex> cf, int dest) {
	int reset=0x80;

	if (cf.size() >= 2048) {
		//printf ("error: set filter tabs %i dest: %i\n",(int)cf.size(),dest);
		//assert(0);
		return;
	}

	if (!setFilterReady) {
		assert(dest<2);
		if (dest>1)
			return;
		qDebug() << "filter not ready, saved" << dest;
		next_coef[dest] = cf;
		return;
	}
	

#ifdef CREATE_MIF
	QFile *file[8];
	if (dest==FILTER_RX) {
		file[7]=new QFile("coefL8A.mif");
		file[6]=new QFile("coefL8B.mif");
		file[5]=new QFile("coefL8C.mif");
		file[4]=new QFile("coefL8D.mif");
		file[3]=new QFile("coefL8E.mif");
		file[2]=new QFile("coefL8F.mif");
		file[1]=new QFile("coefL8G.mif");
		file[0]=new QFile("coefL8H.mif");
		for (int i=0;i<8;i++) {
			if (!file[i]->open(QIODevice::WriteOnly | QIODevice::Text)) {
				qDebug() << "error open file";

			}
			QTextStream out(file[i]);
			out << "DEPTH = 256;\n";
			out << "WIDTH = 36;\n";
			out << "ADDRESS_RADIX = HEX;\n";
			out << "DATA_RADIX = HEX;\n";
			out << "CONTENT\n";
			out << "BEGIN\n";
		}
	} else {
		file[0]=new QFile("coefI8.mif");
		if (!file[0]->open(QIODevice::WriteOnly | QIODevice::Text)) {
			qDebug() << "error open file";
		}
		QTextStream out(file[0]);
		out << "DEPTH = 512;\n";
		out << "WIDTH = 36;\n";
		out << "ADDRESS_RADIX = HEX;\n";
		out << "DATA_RADIX = HEX;\n";
		out << "CONTENT\n";
		out << "BEGIN\n";
	}


#endif

	qDebug() << "set filter tabs "<< cf.size() << " dest:" << dest;
	muteTimer->stop();
	timer->stop();
	//qDebug() << "timer stop";
	dsp->muteSignal(true);

	if (cf.size() == 0)
		return;

	coefPadding = 0;
	if ((cf.size() % 8) > 4)
		coefPadding++;
	else
		if ((cf.size() % 8) > 0)
			coefPadding++;
	if (dest == FILTER_RX)
		coefPadding += 4*8 / COEF_CNT; // last 24 coef have to be zero 

	coef = cf;
	taps = 0;
	cnt = 0;
	it = coef.begin();

	setFilterReady = false;
	setNextFilter(dest, reset);
}

void Hiqsdr::setNextFilter(int dest, int reset) {
	double re,im;
	int32_t ire,iim;
	uint64_t tmp;
	double sum_re=0,sum_im=0;
	int port;

	port = HIQSDR_CTL_PORT;
	filterTimer->stop();

	memset(firBuf, 0, sizeof(firBuf));
	if (dest==FILTER_RX)
		strncpy((char*)firBuf, "Fr", 2);
	else
		strncpy((char*)firBuf, "Ft", 2);

	while (it != coef.end() && ++it!=coef.end()) {

		re = (it->real()*((1<<17)-1));
		im = (it->imag()*((1<<17)-1));
		sum_re += it->real();
		sum_im += it->imag();

		if (re>0) re+= 0.5;
		if (re<0) re-= 0.5;
		if (im>0) im+= 0.5;
		if (im<0) im-= 0.5;

		ire = re;
		iim = im;
		tmp =  iim & 0x3ffff;
		tmp = (tmp << 18);
		tmp |= ire & 0x3ffff;
		// printf ("%i %05lx,  %f + %f i\n",taps, tmp,  it->real(),it->imag());

		firBuf[2+cnt*COEF_LEN+0] = tmp >>  0;
		firBuf[2+cnt*COEF_LEN+1] = tmp >>  8;
		firBuf[2+cnt*COEF_LEN+2] = tmp >> 16;
		firBuf[2+cnt*COEF_LEN+3] = tmp >> 24;
		firBuf[2+cnt*COEF_LEN+4] = tmp >> 32 | reset;

#ifdef CREATE_MIF
		if (dest==FILTER_RX) {

			QTextStream out(file[taps & 7]);
			char s[64];
			sprintf (s,"%03x : %05x;\n",taps/8,tmp);
			out << s;
		} else {
			QTextStream out(file[0]);
			char s[64];
			sprintf (s,"%03x : %05x;\n",taps,tmp);
			out << s;
		}
#endif
		//printf ("%i %i %f %f %u %llx\n",cnt, taps, it->real(),re, ire, tmp);
		reset = 0;
		cnt++;
		taps++;
		if (cnt==COEF_CNT) {
			ctlSocket->writeDatagram(QByteArray((const char*)firBuf, 2+COEF_CNT*COEF_LEN), QHostAddress(hiqsdrIP), port);
			filterTimer->start(FILTER_TIMEOUT);
			cnt = 0;
			return;
		}
	}
	if (it==coef.end()) {

		if (cnt<COEF_CNT) {
			// handle last partialy filled buffer
			ctlSocket->writeDatagram(QByteArray((const char*)firBuf, 2+COEF_CNT*COEF_LEN), QHostAddress(hiqsdrIP), port);
			filterTimer->start(FILTER_TIMEOUT);
			taps+=(COEF_CNT-cnt);
			cnt = COEF_CNT;
			return;
		}

		if (coefPadding>0) {
			ctlSocket->writeDatagram(QByteArray((const char*)firBuf, 2+COEF_CNT*COEF_LEN), QHostAddress(hiqsdrIP), port);
			filterTimer->start(FILTER_TIMEOUT);
			coefPadding--;
			return;
		}

		// printf ("tabs %i, sum re %f sum im %f\n",taps, sum_re, sum_im);
		// ready 
#ifdef CREATE_MIF
		if (dest==FILTER_RX) {
			for (int i=0;i<8;i++)  {
				QTextStream out(file[i]);
				out << "END;\n";
				file[i]->close();
				delete file[i];
			}
		} else {
			QTextStream out(file[0]);
			out << "END;\n";
			file[0]->close();
			delete file[0];
		}
#endif
		setFilterReady = true;
		qDebug() << "filter ready";
		for (int i=0;i<2;i++) {
			if (next_coef[i].size()>0) {
				setFilter(next_coef[i], i);
				next_coef[i].clear();
				return;
			}
		}
		muteTimer->start(200);
		return;
	}
}

void Hiqsdr::setFFT(int time, int size, int src) {
	//qDebug() << "timer stop" << (time>0);
	if (time>0)
		timer->start(time);
	else 
		timer->stop();
	fftTime = time;
	rawSrc = src;
	fftSize = size;
}

int Hiqsdr::diffStep(unsigned int reg, int val) {
	if (reg>=MAX_ENCODERS)
		return 0;

	if (initStep[reg]) {
		lastStep[reg] = val;
		initStep[reg] = false;
		return 0;
	}
	int st = val-lastStep[reg];
	//qDebug() <<  reg << st; 

	if (val-lastStep[reg] > 0x8000)
		st = val-lastStep[reg]-0x10000;
	else if (val-lastStep[reg] < -0x8000)
		st = val-lastStep[reg]+0x10000;
	if (abs(st)>40)
		st*=4;
	lastStep[reg] = val;
	//qDebug() <<  st; 
	return st;
}

void Hiqsdr::readCtlHiqsdr() {
	QByteArray datagram;
	QHostAddress sender;
	quint16 senderPort;
	int len;
	uint8_t buf[64];
	double f;

	while (ctlSocket->hasPendingDatagrams()>0) {
		datagram.resize(ctlSocket->pendingDatagramSize());
		len = ctlSocket->readDatagram((char*)buf, sizeof(buf), &sender, &senderPort);

		if (len == 22 && buf[0]=='Q' && buf[1]=='s') {

			if (memcmp(oldCtlBuf+2, buf+2,4)!=0) {
				f = buf[2] + (buf[3]<<8) + (buf[4]<<16) + (buf[5]<<24); 
				emit controlCommand (SRC_SDR, CMD_RX_FREQ, (int)(f*hiqsdrClock/0x100000000LL+0.5));
			}
			if (memcmp(oldCtlBuf+6, buf+6,4)!=0) {
				f = buf[6] + (buf[7]<<8) + (buf[8]<<16) + (buf[9]<<24); 
				emit controlCommand (SRC_SDR, CMD_TX_FREQ, (int)(f*hiqsdrClock/0x100000000LL+0.5));
			}

			if (oldCtlBuf[12]!=buf[12]) {
				int div1=8,div2=40;
				switch (buf[12]>>6) {
					case 0:
						div1 = 8;
						break;
					case 1:
					case 3:
						div1 = 2;
						break;
					case 2:
						div1 = 40;
						break;

				}
				div2 = (buf[12] & 0x3f) + 1;
				emit controlCommand (SRC_SDR, CMD_DIV1, div1);
				emit controlCommand (SRC_SDR, CMD_DIV2, div2);
			}
			memcpy(oldCtlBuf, buf, len);
		} 
		else if (len == 24 && buf[0]=='S' && buf[1]=='t') {
			// echos
		}
		else if (len == 2 && buf[0]=='F' && buf[1]=='r') {
			setNextFilter(FILTER_RX, 0);
		}
		else if (len == 2 && buf[0]=='F' && buf[1]=='t') {
			setNextFilter(FILTER_TX, 0);
		} 
		else if (len == 5 && buf[0]=='I' && buf[1]=='O') {
			int val = (buf[3]<<8) + buf[4];
			int reg = buf[2];
			// qDebug() << "IO Packet" <<  reg << val;
			switch (reg) {
				case 0 ... 7:
					emit controlCommand (SRC_SDR, CMD_ENCODER0+reg, diffStep(reg, val));
					break;
				case 8:
					emit controlCommand (SRC_SDR, CMD_VOLUME, val);
					break;
			}
		}
		else {
			qDebug() << "unknown Packet" << len << (char)buf[0] << (char)buf[1];
		}

	}
}

void Hiqsdr::writeHiqsdr(QByteArray buf) {

	const short *in = (const short*)buf.data(); 
	int itemsToSend = buf.size()/sizeof(short); 
	int inLevel=0;

	if (audioOff)
		return;

	if (!active)
		return;

	for (int i=0;i<itemsToSend;i++) {
		if (toneTest)
			inLevel = dsp->mod(dsp->tone(), udpBuf+txPos);
		else
			inLevel = dsp->mod(*in, udpBuf+txPos);
		// printf("%x\n",*in);
		txPos += 4;
		in++;
		if (txPos >= (int)sizeof(udpBuf)) {
			//qDebug() << txPos;
			txSocket->writeDatagram(QByteArray((const char*)udpBuf, txPos), QHostAddress(hiqsdrIP), HIQSDR_TX_PORT);
			txPos = 2;
		}
	}
	emit controlCommand (SRC_SDR, CMD_IN_LEVEL, inLevel);
}

void Hiqsdr::writeHiqsdrIQ(QByteArray buf) {

	if (!active)
		return;
	txBufList.append(buf);
}

void Hiqsdr::readHiqsdr() {
	int i;
	QHostAddress sender;
	quint16 senderPort;
	int len;
	static unsigned char seqNr =0;

	while (rxSocket->hasPendingDatagrams()>0) {

		len = rxSocket->readDatagram((char*)d_temp_buff, sizeof(d_temp_buff), &sender, &senderPort);
		if (len != PACKETSIZE_RX) {
			PDEBUG(ERR1,"unexpected rx packet %i", len);
			continue;
		}

		if ((lastTX ^ d_temp_buff[1]) & 2) 
			emit controlCommand (SRC_SDR, CMD_ADC_OV, (d_temp_buff[1] & 2) == 2);
		if ((lastTX ^  d_temp_buff[1]) & 1) {
			emit controlCommand (SRC_SDR, CMD_TX, (d_temp_buff[1] & 1) == 1);
			ptt = d_temp_buff[1] & 1;
			if (!fdx) {
				if (ptt)
					dsp->muteSignal(true);
				else
					muteTimer->start(200);
			}
		}
		lastTX = d_temp_buff[1];

		if (audioOff) {
			emit audioRXRaw(QByteArray((char*)d_temp_buff, len));
			if (txBufList.size()>0)
				txSocket->writeDatagram(txBufList.takeFirst(), QHostAddress(hiqsdrIP), HIQSDR_TX_PORT);
			continue;
		}

		if (++seqNr != (unsigned char)d_temp_buff[0])
			printf("%i: lost packets %u\n",seqNr, abs((unsigned char)d_temp_buff[0]-seqNr));
		seqNr =  (unsigned char)d_temp_buff[0];
		len-=2;	// skip header
		len /= 6;
		int bi = 2;

		dsp->bufReset();
		for (i=0;i < len; i++) {
			int re = (d_temp_buff[bi+0]<<8) + (d_temp_buff[bi+1]<<16) + (d_temp_buff[bi+2]<<24);
			// printf("%02x:%02x:%02x\n",d_temp_buff[bi+0],d_temp_buff[bi+1],d_temp_buff[bi+2]);
			re = re >> 8;
			double ref = (double) re * (1.0 / (double) ((1LL << (NBITS-1)) - 1));

			int im = (d_temp_buff[bi+3]<<8) + (d_temp_buff[bi+4]<<16) + (d_temp_buff[bi+5]<<24);
			im = im >> 8;
			double imf = (double) im * (1.0 / (double) ((1LL << (NBITS-1)) - 1));

			dsp->rxAdd(ref, imf);

			bi += 6;
		}
		emit audioRX (QByteArray((char*)dsp->outBuf, dsp->bufPos*sizeof(short)));
	}
}

void Hiqsdr::agcTimeout() {
	double v  = dsp->getSmtr();
	if (v==0) 
		emit controlCommand (SRC_SDR, CMD_RSSI, 0);
	else 
		// dB * 1000	
		emit controlCommand (SRC_SDR, CMD_RSSI, (int)((log10(v)*10+49.6+smtrOffset)*1000));
}

void Hiqsdr::timeout() {
	if (!active)
		return;

	if ((fftSize>0) && rawRead) {
		PDEBUG(ERR1,"start");
		rawAddr = 0;
		switch (rawSrc) {
			case RAW_SRC_FULL:
				rawHiqSocket->writeDatagram("bb", 2, QHostAddress(hiqsdrIP), HIQSDR_RAW_PORT);
				break;
			case RAW_SRC_CIC_A:
				rawHiqSocket->writeDatagram("cc", 2, QHostAddress(hiqsdrIP), HIQSDR_RAW_PORT);
				break;
			case RAW_SRC_CIC_B:
				rawHiqSocket->writeDatagram("dd", 2, QHostAddress(hiqsdrIP), HIQSDR_RAW_PORT);
				break;
		}
		rawRead = false;
		QTimer::singleShot(fftTime*3/4, this, SLOT(reReadRaw()));
	}
	ctlSocket->writeDatagram("Qs", 2, QHostAddress(hiqsdrIP), HIQSDR_CTL_PORT);
}

void Hiqsdr::readRaw() {
	char addr[2];
	int ret,av=0;
	QByteArray datagram;
	QHostAddress sender;
	quint16 senderPort;
	static int cc=0;

	fftTimer->stop();
	if (!active)
		return;

	while (rawHiqSocket->hasPendingDatagrams()>0) {
		datagram.resize(rawHiqSocket->pendingDatagramSize());

		ret = rawHiqSocket->readDatagram(datagram.data(), datagram.size(),
				&sender, &senderPort);
		// check ram ram fill size
		if ((fftSize>0) && (rawAddr==0) && !(datagram.data()[1] & 1)) {
			addr[1] = rawAddr >> 8;
			addr[0] = rawAddr & 0xff;
			rawHiqSocket->readAll();
			QTimer::singleShot(20, this, SLOT(reReadRaw()));
			PDEBUG(ERR1,"to fast %i",fftTime);
			return;
		}

		rawData.append(datagram.mid(2));

		rawAddr += (ret-2)/2;

		if (rawAddr >= fftSize*2) {
			for (int i=0;i<128;i++) {
				cc++;
				short t1 = *(short*)(rawData.data()+(i*4+0));
				short t2 = *(short*)(rawData.data()+(i*4+2));
				av+=t1*t1+t2*t2;
			}
			av = sqrt(av/128);
			// setNBLevel(av*6);
			//qDebug() << av  << nbLevel;
			emit fftData(rawData);
			rawData.clear();
			rawRead = true;
			return;
		}
	}

	addr[1] = rawAddr >> 8;
	addr[0] = rawAddr & 0xff;
	ret = rawHiqSocket->writeDatagram(addr ,2, QHostAddress(hiqsdrIP), HIQSDR_RAW_PORT);
	if (ret != 2)
		PDEBUG(ERR1,"error write raw addr");
	fftTimer->start(FFT_TIMEOUT);
}

void Hiqsdr::reReadRaw() {
	char addr[2];
	addr[1] = rawAddr >> 8;
	addr[0] = rawAddr & 0xff;
	rawHiqSocket->writeDatagram(addr ,2, QHostAddress(hiqsdrIP), HIQSDR_RAW_PORT);
}

void Hiqsdr::muteTimeout() {
	dsp->muteSignal(false);
	muteTimer->stop();
	if (fftTime>0)
		timer->start(fftTime);
	//qDebug() << "timer start" << fftTime;
}

void Hiqsdr::filterTimeout() {
	PDEBUG(ERR1,"error: filter timeout");
	setFilterReady = true;
}

void Hiqsdr::fftTimeout() {
	PDEBUG(ERR1,"error: fft timeout");
	rawRead = true;
}

void Hiqsdr::setMode(int m) {
	dsp->setMode(m);
	setCWMode(m == MODE_CW);
}

void Hiqsdr::setGain(double m) {
	dsp->setGain(m);
}

void Hiqsdr::setAGCDec(double m) {
	dsp->setAGCDec(m);
}

void Hiqsdr::setMicGain(double m) {
	dsp->setMicGain(m);
}

void Hiqsdr::setVolume(double m) {
	dsp->setVolume(m);
}

void Hiqsdr::setToneTest(bool on, double f1, double level1, double f2, double level2) {
	toneTest = on;
	dsp->setTestTone(f1, level1, f2, level2);
}

void Hiqsdr::setActive(bool on) {
	active = on;
}

void Hiqsdr::setAudioOff(bool on) {
	audioOff = on;
}

void Hiqsdr::setNotch(int v) {
	dsp->setNotch(v);
}

void Hiqsdr::setSquelch(int v) {
	dsp->setSquelch(v);
}

void Hiqsdr::terminate() {
    hiqRun = false;
}

void Hiqsdr::run() {

    while(hiqRun) {
		sleep(1);
	}
}


